/*  -*- tab-width: 4 -*-  */


#include <sys/time.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include "header.h"


#define VBUFSIZE 100


#define ISDATACOMMAND(line) (strcasecmp((line), "data\r\n") == 0 || \
							 strcasecmp((line), "data\n") == 0)
#define ISENDOFHEADER(line) (strcmp((line), "\r\n") == 0 || \
							 strcmp((line), "\n") == 0)
#define ISENDOFDATA(line)   (strcmp((line), ".\r\n") == 0 || \
							 strcmp((line), ".\n") == 0)
#define ISLNL(p) ((p) == '\n' && !isspace(*(&(p)+1))) /* is logical newline */


static unsigned char *ptr;
static unsigned char *data;
static unsigned char *eoh;		/* end of header */
static unsigned char *designatorp;
static unsigned char *mime_version;
static unsigned char *content_type, *end_content_type, *end_type_subtype;
static unsigned char *charset, *end_charset;
static unsigned char *content_transfer_encoding,
	*end_content_transfer_encoding;
static int ct, cte;				/* {C-T, C-T-E} field has been written */
static int eightbit_header = 0;


void write_or_store _PRO((FILE *fp, char *buf, size_t count));
void write_or_store_a_line _PRO((char *s, FILE *fp, int flush));
void conv_copy _PRO((FILE *fp, int fd, size_t len));
void parse_header _PRO(());
void rewrite_header _PRO((unsigned char *l, unsigned char *end_l, FILE *fp));
unsigned char *eat_space _PRO(());
unsigned char *eat_field_name _PRO(());
unsigned char *eat_upto_lnl _PRO(());
unsigned char *memlnl _PRO((unsigned char *p));
void b_encode _PRO((unsigned char *m, unsigned char *end, FILE *fp));


/*
 *  Tells sendmail what is heard from MUA and vice versa.
 *  Encodes MUA-to-sendmail data according to RFC 1557.
 */
int filter(stdinfd, stdoutfd, stderrfd)
int stdinfd, stdoutfd, stderrfd;
{
	fd_set s, s1;
	int not_end = 1;
	int maxfdp1 = MAX(STDIN_FILENO, MAX(stdoutfd, stderrfd)) + 1;
	FILE *stdinp;				/* fed into sendmail's stdin */

	if ((stdinp = fdopen(stdinfd, "w")) == NULL)
		perror_exit(1);

	FD_ZERO(&s1);
	FD_SET(STDIN_FILENO, &s1);
	FD_SET(stdoutfd, &s1);
	FD_SET(stderrfd, &s1);
	
	do {
		char buffer[BUFFERSIZE];
		ssize_t n;
		
		memcpy(&s, &s1, sizeof(s1));

		if (select(maxfdp1, &s, NULL, NULL, NULL) < 0)
			perror_exit(1);

		if (FD_ISSET(STDIN_FILENO, &s)) {
			if ((n = read(STDIN_FILENO, buffer, sizeof(buffer))) < 0)
				perror_exit(1);
#ifdef DEBUG
			fwrite(buffer, 1, n, ep);
#endif
			write_or_store(stdinp, buffer, n);

			if (n == 0) {		/* EOF */
				FD_CLR(STDIN_FILENO, &s1);
				fclose(stdinp);
			}
		}

		if (FD_ISSET(stdoutfd, &s)) {
			if ((n = read(stdoutfd, buffer, sizeof(buffer))) < 0)
				perror_exit(1);

			if (n == 0)	/* EOF */
				not_end = 0;
			else {
				writen(STDOUT_FILENO, buffer, n);
#ifdef DEBUG
				fwrite(buffer, 1, n, ep);
#endif
			}
		}
		
		if (FD_ISSET(stderrfd, &s)) {
			if ((n = read(stderrfd, buffer, sizeof(buffer))) < 0)
				perror_exit(1);

			writen(STDERR_FILENO, buffer, n);
#ifdef DEBUG
			fwrite(buffer, 1, n, ep);
#endif
		}
	} while(not_end);

	return 0;
}


/*
 *  Gathers a line from buf and calls write_or_store_a_line().
 */
void write_or_store(fp, buf, count)
FILE *fp;
register char *buf;
size_t count;
{
	static unsigned char *vbuf = NULL;
	static size_t svbuf = 0;
	static unsigned char *p = NULL;
	static enum { ASCII, KOR1, KOR2 } whatkind = ASCII;

	if (vbuf == NULL)			/* the 1st allocation of buffer */
		if ((p = vbuf = malloc(VBUFSIZE)) == NULL)
			perror_exit(1);
		else
			svbuf = VBUFSIZE;

	if (count == 0) { 			/* flush vbuf */
		*p = '\0';
		write_or_store_a_line(vbuf, fp, 1);
		p = vbuf;
		whatkind = ASCII;
	}
	
	while (count > 0) {
		int endofline = 0;
		
		if (p - vbuf >= svbuf-2) { /* vbuf full */
			size_t n = p - vbuf;

			if (svbuf + svbuf/2 <= MAXLINELEN) {
				/* expand vbuf */			
				if ((vbuf = realloc(vbuf, svbuf += svbuf/2)) == NULL)
					perror_exit(1);
				p = vbuf + n;
#ifdef DEBUG
				fprintf(ep, "vbuf expanded to %d bytes.\n", svbuf);
#endif
			}
			else {				/* cannot expand any more */
				if (whatkind == KOR1) {
					*p++ = *buf++;
					count--;
					whatkind = KOR2;
				}
				endofline = 1; /* possible mail bomb? */
			}
		}

		if (!endofline) {
			switch (whatkind) {
			case KOR2:
				if (isksc(*buf)) whatkind = KOR1;
				else whatkind = ASCII; break;
			case ASCII:
				if (isksc(*buf)) whatkind = KOR1; break;
			case KOR1:
				whatkind = KOR2; break;
			}
			if ((*p++ = *buf++) == '\n')
				endofline = 1;
			count--;
		}
		
		if (endofline) {
			*p = '\0';
			write_or_store_a_line(vbuf, fp, 0);
			p = vbuf;
			whatkind = ASCII;
		}
	}

	fflush(fp);
}


void write_or_store_a_line(s, fp, flush)
char *s;
FILE *fp;
int flush;
{
	static enum { COMMAND, DATA } state = COMMAND;
	
	if (use_smtp)
		switch (state) {
		case COMMAND:
			fputs(s, fp);
			if (ISDATACOMMAND(s)) {
				state = DATA;
				fseek(tmpp, 0L, SEEK_SET);
			}
			break;
		case DATA:
			if (fputs(s, tmpp) == EOF) perror_exit(1);
			if (ISENDOFDATA(s)) {
				state = COMMAND;
				if (putc(0, tmpp) == EOF) perror_exit(1);
				if (fflush(tmpp) == EOF) perror_exit(1);
#ifdef DEBUG
				conv_copy(op, tmp, ftell(tmpp));
#endif
				conv_copy(fp, tmp, ftell(tmpp));
			}
			break;
		}
	else {						/* non-smtp mode */
		if (fputs(s, tmpp) == EOF) perror_exit(1);

		if (flush) {
			if (putc(0, tmpp) == EOF) perror_exit(1);
			if (fflush(tmpp) == EOF) perror_exit(1);
#ifdef DEBUG
			conv_copy(op, tmp, ftell(tmpp));
#endif
			conv_copy(fp, tmp, ftell(tmpp));
		}
	}
}


void conv_copy(fp, fd, len)
FILE *fp;
int fd;
size_t len;						/* length including terminating zero */
{
	designatorp = NULL;
	eightbit_header = 0;
	content_type = NULL;
	charset = NULL;
	content_transfer_encoding = NULL;
	mime_version = NULL;

	if ((data = mmap(0, len, PROT_READ, MAP_SHARED, fd, 0)) == -1)
		perror_exit(1);

	if ((eoh = strstr(data, "\n\n")) == NULL &&
		(eoh = strstr(data, "\n\r\n")) == NULL)
		eoh = data + len - 1;
	else
		eoh++;

	for (ptr = eoh; *ptr; ptr++)
		if (isksc(*ptr)) {
			while (*--ptr != '\n')
				;
			designatorp = ptr + 1;
			break;
		}

	ptr = data;
	
	parse_header();

	if (designatorp && isin(charset, end_charset - charset, charsets)) {
		/* body has an 8bit char and charset is in charsets */
		unsigned char *bol;

		ct = 0; cte = 0;
		
		/* write header */
		for (bol = ptr = data; ptr < eoh; bol = ptr) {
			ptr = memlnl(bol) + 1;
			rewrite_header(bol, ptr, fp);
		}
		if (mime_version == NULL) {
			fputs("MIME-Version: 1.0 (converted by kesmpre V0.2)", fp);
			putnl(fp);
		}
		if (!ct) {
			fputs("Content-Type: text/plain; charset=iso-2022-kr", fp);
			putnl(fp);
		}
		if (!cte) {
			fputs("Content-Transfer-Encoding: 7bit", fp);
			putnl(fp);
		}
		
		/* write body */
		while (ptr < designatorp)
			putc(*ptr++, fp);

		fprintf(fp, "\033$)C"); /* put out the designator */
#ifdef NLAFTERDESIG
		putnl(out);
#endif

		for (bol = ptr; *ptr; ptr++)
			if (*ptr == '\n') {
				convert(bol, fp);
				bol = ptr+1;
			}

		if (ptr > bol)
			convert(bol, fp);
	}
	else {						/* body isn't to be encoded */
		unsigned char *bol = data; /* begin of line */
		
		for (ptr = data; ptr < eoh; ptr++)
			if (*ptr == '\n') {
				b_encode(bol, ptr+1, fp);
				bol = ptr+1;
			}

		if (ptr > bol)
			b_encode(bol, ptr, fp);

		while (*ptr)
			putc(*ptr++, fp);
	}

	munmap(data, len);
}

void parse_header()
{
	while (ptr < eoh) {
		char *b, *e;

		b = ptr; eat_field_name(); e = ptr; eat_space();
		if (*ptr == ':') {
			if (memstrcasecmp(b, e - b, "Content-Type") == 0) {
				ptr++; eat_space();
				content_type = ptr; end_content_type = memlnl(ptr);

				while (ptr < end_content_type && *ptr != ';')
					ptr++;
				end_type_subtype = ptr;
				
				if ((b = memcasestr(content_type, end_content_type -
									content_type, "charset")) != NULL) {
					ptr = b + strlen("charset"); eat_space();
					if (*ptr == '=') {
						ptr++; eat_space();
						charset = ptr;
						while (ptr < end_content_type &&
							   *ptr != ';' && !isspace(*ptr))
							ptr++;
						end_charset = ptr;
					}
				}
			}				
			else if (memstrcasecmp(b, e - b, "Content-Transfer-Encoding")
					 == 0) {
				ptr++; eat_space();
				content_transfer_encoding = ptr;
				while (ptr < eoh && !isspace(*ptr))
					ptr++;
				end_content_transfer_encoding = ptr;
			}
			else if (memstrcasecmp(b, e - b, "MIME-Version") == 0)
				mime_version = b;
		}
		eat_upto_lnl();
	}
}
	
unsigned char *eat_space()
{
	while (ptr < eoh && isspace(*ptr) && !ISLNL(*ptr))
		ptr++;

	return ptr;
}

unsigned char *memlnl(p)
unsigned char *p;
{
	while (p < eoh && !ISLNL(*p))
		p++;

	if (p == eoh)
		if (*(eoh-1) == '\n')
			p = eoh-1;
	
	return p;
}

unsigned char *eat_upto(c)
int c;
{
	while (ptr < eoh && *ptr != c)
		ptr++;
	if (ptr < eoh)
		ptr++;

	return ptr;
}

unsigned char *eat_upto_lnl()
{
	ptr = memlnl(ptr);
	
	if (ptr < eoh)
		ptr++;

	return ptr;
}

unsigned char *eat_field_name()
{
	while (ptr < eoh && !iscntrl(*ptr) && !isspace(*ptr) && *ptr != ':')
		ptr++;

	return ptr;
}


void rewrite_header(l, end_line, fp)
unsigned char *l;				/* logical line */
unsigned char *end_line;		/* end of logical line */
FILE *fp;
{
	if (content_type && l <= content_type && content_type < end_line) {
	    fwrite(l, 1, content_type - l, fp); l = content_type;
		
		if (memchr(l, '/', end_type_subtype - l) == NULL)
			fputs("text/plain", fp);
		else
			fwrite(l, 1, end_type_subtype - l, fp);
		l = end_type_subtype;

		if (charset) {
			fwrite(l, 1, charset - l, fp);
			fputs("iso-2022-kr", fp);
			l = end_charset;
		}
		else {
			unsigned char *l1;

			for (l1 = end_content_type-1; isspace(*l1); l1--)
				;
			fwrite(l, 1, ++l1 - l, fp);
			fputs("; charset=iso-2022-kr", fp);
			l = l1;
		}
		fwrite(l, 1, end_line - l, fp); l = end_line;

		ct = 1;
	}
	else if (content_transfer_encoding && l <= content_transfer_encoding &&
			 content_transfer_encoding < end_line) {
		fwrite(l, 1, content_transfer_encoding - l, fp);
		l = content_transfer_encoding;

		if (memstrcasecmp(l, end_content_transfer_encoding-l, "8bit") == 0) {
			fputs("7bit", fp);
			l = end_content_transfer_encoding;
		}
		else {
			fwrite(l, 1, end_content_transfer_encoding - l, fp);
			l = end_content_transfer_encoding;
		}
		fwrite(l, 1, end_line - l, fp); l = end_line;

		cte = 1;
	}
	else {
		unsigned char *bol;
		
		for (bol = l; l < end_line; l++)
			if (*l == '\n') {
				b_encode(bol, l+1, fp);
				bol = l+1;
			}

		if (l > bol)
			b_encode(bol, l, fp);
	}
}

void b_encode(m, end, fp)
unsigned char *m, *end;
FILE *fp;
{
	unsigned char *tptr;
	
	while (m < end) {
		while (m < end && *m < 0x80)
			putc(*m++, fp);		/* write ASCIIchars without conversion */
		
		tptr = m;
		while (m < end && (*m >= 0x80 || *m == ' '))
			m++;				/* gather a Korean chunk */
		if (m != tptr) {
			fprintf(fp, "=?EUC-KR?B?");
			string_to_base64(fp, tptr, m - tptr);
			fprintf(fp, "?=");
		}
	}
}
