/* -*- tab-width: 4  -*- */

/*
 *  Refer to Richard Stevens' Advanced Programming in the UNIX Environment
 *  for the original code which I have copied.
 */


#include <unistd.h>

#include "header.h"


pid_t invoke_real_sendmail(stdinfd, stdoutfd, stderrfd, argv)
int *stdinfd, *stdoutfd, *stderrfd;
char *argv[];
{
	int fd1[2], fd2[2], fd3[2];
	pid_t pid;

	if (pipe(fd1) < 0 || pipe(fd2) < 0 || pipe(fd3) < 0)
		perror_exit(1);

	if ((pid = fork()) < 0)
		perror_exit(1);

	if (pid == 0) {				/* child */
		close(fd1[1]);
		close(fd2[0]);
		close(fd3[0]);

		if (fd1[0] != STDIN_FILENO) {
			if (dup2(fd1[0], STDIN_FILENO) != STDIN_FILENO)
				perror_exit(1);
			close(fd1[0]);
		}
		if (fd2[1] != STDOUT_FILENO) {
			if (dup2(fd2[1], STDOUT_FILENO) != STDOUT_FILENO)
				perror_exit(1);
			close(fd2[1]);
		}
		if (fd3[1] != STDERR_FILENO) {
			if (dup2(fd3[1], STDERR_FILENO) != STDERR_FILENO)
				perror_exit(1);
			close(fd3[1]);
		}

		if (execvp(SENDMAIL, argv) < 0)
			perror_exit(1);
	}

	/* parent */

	close(fd1[0]);
	close(fd2[1]);
	close(fd3[1]);

	*stdinfd = fd1[1];
	*stdoutfd = fd2[0];
	*stderrfd = fd3[0];

	return pid;
}
