/* -*- tab-width: 4 -*- */


#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <ctype.h>

#include "header.h"


#define CFNAMELEN   512
#define CHARSETLEN  64


#define ISSEPARATOR(c) ((c) == ':' || isspace(c))


#ifdef DEBUG
FILE *ep;						/* log file */
FILE *op;						/* output log */
#endif

int use_smtp = 0;				/* 1 if -bs option is given */
/* charsets to regard as EUC-KR */
char *charsets[MAXCHARSETS+1] = {
	"EUC-KR",
	"ISO-2022-KR",
	"ISO-8859-1",
	"NONE",
	NULL
};


void read_config();


int main(argc, argv)
int argc;
char *argv[];
{
	int stdinfd, stdoutfd, stderrfd; /* stdin, stdout, stderr of child */
	pid_t pid;
	char **a;

#ifdef DEBUG
	{
		char **a = argv;
		
		ep = fopen("kesmpre.log", "a"); setbuf(ep, NULL);
		op = fopen("kesmpre.out", "a"); setbuf(op, NULL);
		fprintf(ep, "Invoked with argv:");
		while (*a) fprintf(ep, " %s", *a++);
		putc('\n', ep);
	}
#endif
	
	/* see sendmail(8) to know what -bs is. */
	for (a = argv+1; *a; a++)
		if (strcmp(*a, "-bs") == 0)
			use_smtp = 1;

	read_config();

#ifdef DEBUG
	{
		char **c;
		
		for (c = charsets; *c; c++)
			fprintf(ep, "charset to encode: %s\n", *c);
	}
#endif
	
	/* temporary file will be automatically removed on exit. */
	if ((tmp = maketmpfile()) < 0)
		perror_exit(1);
	if ((tmpp = fdopen(tmp, "w")) == NULL)
		perror_exit(1);
	
	pid = invoke_real_sendmail(&stdinfd, &stdoutfd, &stderrfd, argv);

	filter(stdinfd, stdoutfd, stderrfd);

#ifdef DEBUG
	fclose(ep);
	fclose(op);
#endif
	
	return 0;
}


void read_config()
{
	char *env, *e, *p;
	char cfname[CFNAMELEN];
	int fd = -1;
	struct stat s;
	int i;
	
	if ((env = getenv("KESMPRE")) != NULL) {
		e = env + strlen(env);
	}
	else {
		strncpy(cfname, getenv("HOME"), CFNAMELEN - 10);
		cfname[CFNAMELEN - 10] = '\0';
		strcat(cfname, "/.kesmpre");
		
		if ((fd = open(cfname, O_RDONLY)) < 0)
			if ((fd = open(SYSCFNAME, O_RDONLY)) < 0)
				return;
	
		if (fstat(fd, &s) < 0)
			return;
		if ((env = mmap(0, s.st_size, PROT_READ, MAP_SHARED, fd, 0)) == -1)
			return;

		e = env + s.st_size;
	}

	p = env;
	
	for (i = 0; i < MAXCHARSETS; i++) {
		char b[CHARSETLEN];
		char *q = b;
		
		while (p < e && ISSEPARATOR(*p))
			p++;
		if (p == e) break;

		while (p < e && q < b + CHARSETLEN - 1 && !ISSEPARATOR(*p))
			*q++ = *p++;
		*q = '\0';

		charsets[i] = strdup(b);
		
		while (p < e && !ISSEPARATOR(*p))
			p++;
	}

	if (i > 0)					/* if i == 0, use the default. */
		charsets[i] = NULL;

	if (fd >= 0) {
		munmap(env, s.st_size);
		close(fd);
	}
}
